package main

import (
	"context"
	"fmt"
	"net"
	"os"
	"regexp"
	"time"

	"cloud.google.com/go/pubsub"
	"github.com/jmoiron/sqlx"
	account "gitlab.mashroom.com/mashroom/backend/account/api"
	"gitlab.mashroom.com/mashroom/backend/common/api/health"
	"gitlab.mashroom.com/mashroom/backend/common/generic/addr"
	"gitlab.mashroom.com/mashroom/backend/common/generic/env"
	"gitlab.mashroom.com/mashroom/backend/common/generic/grpc"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	"gitlab.mashroom.com/mashroom/backend/common/generic/tracer"
	gmailparser "gitlab.mashroom.com/mashroom/backend/gmailparser/api"
	"gitlab.mashroom.com/mashroom/backend/gmailparser/internal/model/postgres"
	"gitlab.mashroom.com/mashroom/backend/gmailparser/internal/parser"
	"gitlab.mashroom.com/mashroom/backend/gmailparser/internal/service"
	"gitlab.mashroom.com/mashroom/backend/gmailparser/version"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	mixpanel "gitlab.mashroom.com/mashroom/backend/mixpanel/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
	"golang.org/x/oauth2"
	"golang.org/x/oauth2/google"
	admin "google.golang.org/api/admin/directory/v1"
	"google.golang.org/api/gmail/v1"
	"google.golang.org/api/option"
)

var (
	envProjectID      = env.ResolveString("PROJECT_ID", "mashroom-email-listener")
	envSubjectID      = env.ResolveString("SUBJECT_ID", "portals@mashroom.com")
	envServiceKeyFile = env.ResolveString("SERVICE_KEY_FILE", "private.json")
)

var (
	envListingService = listing.NewListingServiceClient(
		grpc.Dial(env.ResolveString("LISTING_SERVICE_URL", "mashroom-listing:10000"), "Listing"),
	)
	envNotificationService = notification.NewNotificationServiceClient(
		grpc.Dial(env.ResolveString("NOTIFICATION_SERVICE_URL", "mashroom-notification:10000"), "Notification"),
	)
	envAccountService = account.NewAccountServiceClient(
		grpc.Dial(env.ResolveString("ACCOUNT_SERVICE_URL", "mashroom-account:10000"), "Account"),
	)
	envMixpanelService = mixpanel.NewMixpanelServiceClient(
		grpc.Dial(env.ResolveString("MIXPANEL_SERVICE_URL", "mashroom-mixpanel:10000"), "Mixpanel"),
	)
)

var (
	envSubscriptionZoopla = env.ResolveString(
		"G_ZOOPLA",
		"notifications/zoopla-test=projects/mashroom-email-listener/topics/testingportals=zoopla-test",
	)
	envSubscriptionRightmove = env.ResolveString(
		"G_RIGHTMOVE",
		"notifications/rightmove-test=projects/mashroom-email-listener/topics/testingportals=rightmove-test",
	)
)

var (
	parseSetRightmove = parser.ParseSet{
		Name: []*regexp.Regexp{
			regexp.MustCompile(`Name:\s*(.*?)[\r\n]`),
		},
		Email: []*regexp.Regexp{
			regexp.MustCompile(`Email:\s*(.*?)[\r\n]`),
		},
		Phone: []*regexp.Regexp{
			regexp.MustCompile(`Telephone:\s*(.*?)[\r\n]`),
		},
		Address: []*regexp.Regexp{
			regexp.MustCompile(`(?s)Property details.*?Address:\s(.*?)[\r\n]`),
		},
		Property: []*regexp.Regexp{
			regexp.MustCompile(`Reference:\s\d+_([a-z0-9-]+)\s*`),
		},
		Reference: []*regexp.Regexp{
			regexp.MustCompile(`(-)`),
		},
	}
	parseSetZoopla = parser.ParseSet{
		Name: []*regexp.Regexp{
			regexp.MustCompile(`Name:\s*(.*?)[\r\n]`),
		},
		Email: []*regexp.Regexp{
			regexp.MustCompile(`Email:\s*(.*?)[\r\n]`),
			regexp.MustCompile(`Email address:\s*(.*?)[\r\n]`),
		},
		Phone: []*regexp.Regexp{
			regexp.MustCompile(`Telephone:\s*(.*?)[\r\n]`),
			regexp.MustCompile(`Telephone number:\s*(.*?)[\r\n]`),
		},
		Address: []*regexp.Regexp{
			regexp.MustCompile(`Address:\s*(.*?)[\r\n]`),
			regexp.MustCompile(`Full address:\s*(.*?)[\r\n]`),
		},
		Property: []*regexp.Regexp{
			regexp.MustCompile(`Your property ref:\s*([a-z0-9-]+)[\r\n]`),
			regexp.MustCompile(`Your property ref:\s*\d+_([a-z0-9-]+)[\r\n]`),
		},
		Reference: []*regexp.Regexp{
			regexp.MustCompile(`(https?://www.zoopla\.co\.uk/.*/details/\d+)`),
		},
	}
)

var (
	envScheduleInterval = env.ResolveDuration(
		"SCHEDULE_INTERVAL",
		time.Minute*5,
	)
)

func main() {
	defer tracer.Close()

	ctx := context.Background()

	conn := fmt.Sprintf(
		"postgres://%s:%s@%s:%s/%s?sslmode=disable",
		os.Getenv("POSTGRES_USERNAME"),
		os.Getenv("POSTGRES_PASSWORD"),
		os.Getenv("POSTGRES_HOST"),
		os.Getenv("POSTGRES_PORT"),
		os.Getenv("POSTGRES_DATABASE"),
	)

	db := sqlx.MustOpen("postgres", conn)

	repo, err := postgres.NewRepository(db)
	if err != nil {
		panic(err)
	}

	err = repo.Migrate(ctx)
	if err != nil {
		panic(err)
	}

	srv, err := service.NewService(&service.Config{
		Repository: repo,
	})
	if err != nil {
		panic(err)
	}

	listener, err := net.Listen("tcp", addr.Grpc)
	if err != nil {
		panic(err)
	}

	grpcServer, err := grpc.NewServer("gmailparser")
	if err != nil {
		panic(err)
	}

	clientsource, err := serviceAccount(
		ctx,
		envSubjectID,
		admin.AdminDirectoryUserReadonlyScope,
		admin.AdminDirectoryGroupReadonlyScope,
		gmail.GmailReadonlyScope,
		gmail.GmailModifyScope,
	)
	if err != nil {
		panic(err)
	}

	gm, err := gmail.NewService(ctx, option.WithTokenSource(clientsource))
	if err != nil {
		panic(err)
	}

	pubsubsource, err := serviceAccount(
		ctx,
		"",
		pubsub.ScopePubSub,
	)
	if err != nil {
		panic(err)
	}

	psub, err := pubsub.NewClient(ctx, envProjectID, option.WithTokenSource(pubsubsource))
	if err != nil {
		panic(err)
	}

	subs := []*parser.Subscription{
		{
			Name:            "rightmove",
			Config:          envSubscriptionRightmove,
			ParseSet:        parseSetRightmove,
			ListingPlatform: mixpanel.ListingPlatform_LISTING_PLATFORM_RIGHTMOVE,
		},
		{
			Name:            "zoopla",
			Config:          envSubscriptionZoopla,
			ParseSet:        parseSetZoopla,
			ListingPlatform: mixpanel.ListingPlatform_LISTING_PLATFORM_ZOOPLA,
		},
	}

	p, err := parser.NewParser(&parser.Config{
		Repository:    repo,
		ProjectID:     envProjectID,
		Subscriptions: subs,
		GmailService:  gm,
		PubsubClient:  psub,
		Connections: parser.Connections{
			Listing:      envListingService,
			Notification: envNotificationService,
			Mixpanel:     envMixpanelService,
			Account:      envAccountService,
		},
		ScheduleInterval: envScheduleInterval,
	})
	if err != nil {
		panic(err)
	}

	err = p.Setup(ctx)
	if err != nil {
		panic(err)
	}

	gmailparser.RegisterGmailparserServiceServer(grpcServer, srv)

	log.Info("gmailparser started")

	healthServer := version.NewHealthServer()
	health.RegisterHealthServiceServer(grpcServer, healthServer)

	_ = grpcServer.Serve(listener)
}

func serviceAccount(ctx context.Context, subject string, scopes ...string) (oauth2.TokenSource, error) {
	bs, err := os.ReadFile(envServiceKeyFile)
	if err != nil {
		return nil, err
	}

	conf, err := google.JWTConfigFromJSON(bs, scopes...)
	if err != nil {
		return nil, err
	}

	if subject != "" {
		conf.Subject = subject
	}

	return conf.TokenSource(ctx), nil
}
