// Package parser provides parser for email enquiries using gmail API
package parser

import (
	"errors"
	"regexp"
	"strings"
	"time"

	"cloud.google.com/go/pubsub"
	account "gitlab.mashroom.com/mashroom/backend/account/api"
	"gitlab.mashroom.com/mashroom/backend/gmailparser/internal/model"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	mixpanel "gitlab.mashroom.com/mashroom/backend/mixpanel/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
	"google.golang.org/api/gmail/v1"
)

// ParseSet set of regexps to extract relevant data from email
type ParseSet struct {
	Name      []*regexp.Regexp
	Email     []*regexp.Regexp
	Phone     []*regexp.Regexp
	Address   []*regexp.Regexp
	Property  []*regexp.Regexp
	Reference []*regexp.Regexp
}

// EmailFields parsed from email body
type EmailFields struct {
	Name      string
	Email     string
	Phone     string
	Address   string
	Property  string
	Reference string
}

// Parse the email body using configrued parse set
func (set *ParseSet) Parse(
	rawContent string,
) (res EmailFields) {
	content := regexp.MustCompile(`=[\r\n]`).ReplaceAllString(rawContent, "")

	res.Name = getOneOfRegexps(content, set.Name)
	res.Email = getOneOfRegexps(content, set.Email)
	res.Phone = getOneOfRegexps(content, set.Phone)
	res.Address = getOneOfRegexps(content, set.Address)
	res.Property = getOneOfRegexps(content, set.Property)
	res.Reference = getOneOfRegexps(content, set.Reference)

	if newPhone := formatPhoneNumber(res.Phone); newPhone != "" {
		res.Phone = newPhone
	}

	return res
}

// Subscription definition
type Subscription struct {
	Name                                string
	Config                              string
	ParseSet                            ParseSet
	ListingPlatform                     mixpanel.ListingPlatform
	label, labelID, topic, subscription string
}

var errInvalidConfig = errors.New("invalid config")

func (sub *Subscription) parseConfig() error {
	es := strings.SplitN(sub.Config, "=", 3)
	if len(es) != 3 {
		return errInvalidConfig
	}

	sub.label = es[0]
	sub.topic = es[1]
	sub.subscription = es[2]

	return nil
}

// Connections of the parser
type Connections struct {
	Listing      listing.ListingServiceClient
	Notification notification.NotificationServiceClient
	Mixpanel     mixpanel.MixpanelServiceClient
	Account      account.AccountServiceClient
}

// Config of the parser
type Config struct {
	Repository       model.Repository
	ProjectID        string
	Subscriptions    []*Subscription
	GmailService     *gmail.Service
	PubsubClient     *pubsub.Client
	Connections      Connections
	ScheduleInterval time.Duration
}

// Parser implementation
type Parser struct {
	Config
	labels map[string]*Subscription
}

// NewParser returns new parser instance
func NewParser(config *Config) (*Parser, error) {
	labels := make(map[string]*Subscription)

	for _, s := range config.Subscriptions {
		err := s.parseConfig()
		if err != nil {
			return nil, err
		}

		labels[s.label] = s
	}

	p := &Parser{
		Config: *config,
		labels: labels,
	}

	return p, nil
}
