package parser

import (
	"context"
	"encoding/base64"
	"encoding/json"
	"errors"
	"io"
	"regexp"
	"strings"
	"sync"
	"time"

	"cloud.google.com/go/pubsub"
	"github.com/nyaruka/phonenumbers"
	account "gitlab.mashroom.com/mashroom/backend/account/api"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcctx"
	"gitlab.mashroom.com/mashroom/backend/common/api/grpcerr"
	"gitlab.mashroom.com/mashroom/backend/common/generic/log"
	"gitlab.mashroom.com/mashroom/backend/gmailparser/internal/model"
	listing "gitlab.mashroom.com/mashroom/backend/listing/api"
	messaging "gitlab.mashroom.com/mashroom/backend/messaging/api"
	mixpanel "gitlab.mashroom.com/mashroom/backend/mixpanel/api"
	notification "gitlab.mashroom.com/mashroom/backend/notification/api"
	"google.golang.org/api/gmail/v1"
)

var (
	errNoProperty = grpcerr.New(grpcerr.InvalidData, "property not found")
	errNoUser     = grpcerr.New(grpcerr.InvalidData, "user not found")
)

const mimeTextPlain = "text/plain"

type notificationPayload struct {
	EmailAddress string `json:"emailAddress"`
	HistoryID    uint64 `json:"historyId"`
}

type message struct {
	ID      string
	From    string
	Subject string
	Body    string
}

func findMessagePartBody(parts []*gmail.MessagePart, bodyData map[string]*string) {
	if parts == nil {
		return
	}

	for _, part := range parts {
		if part.Body != nil && part.Body.Data != "" && part.MimeType != "" {
			bodyData[part.MimeType] = &part.Body.Data
		}

		findMessagePartBody(part.Parts, bodyData)
	}
}

func findMessageBody(msg *gmail.Message) (*string, string) {
	bodyData := make(map[string]*string)

	if msg.Payload != nil {
		if msg.Payload.Body != nil && msg.Payload.Body.Data != "" && msg.Payload.MimeType != "" {
			bodyData[msg.Payload.MimeType] = &msg.Payload.Body.Data
		}

		if msg.Payload.Parts != nil && len(msg.Payload.Parts) > 0 {
			findMessagePartBody(msg.Payload.Parts, bodyData)
		}
	}

	if data, ok := bodyData[mimeTextPlain]; ok {
		return data, mimeTextPlain
	}

	return nil, ""
}

func (p *Parser) loadMessage(id string) (*message, error) {
	msg, err := p.GmailService.Users.Messages.Get("me", id).Format("full").Do()
	if err != nil {
		return nil, err
	}

	var from, subject string

	for _, h := range msg.Payload.Headers {
		switch {
		case strings.EqualFold(h.Name, "from"):
			from = h.Value
		case strings.EqualFold(h.Name, "subject"):
			subject = h.Value
		}
	}

	bodyData, _ := findMessageBody(msg)
	dat := ""

	if bodyData != nil {
		dat = strings.ReplaceAll(strings.ReplaceAll(*bodyData, "-", "+"), "_", "/")
	}

	ss, _ := io.ReadAll(base64.NewDecoder(base64.StdEncoding, strings.NewReader(dat)))

	return &message{
		ID:      id,
		From:    from,
		Subject: subject,
		Body:    string(ss),
	}, nil
}

func (p *Parser) processUnread(
	ctx context.Context,
	entry *log.Entry,
	v *Subscription,
) {
	h, err := p.GmailService.Users.Messages.List("me").LabelIds(v.labelID, "UNREAD").Do()
	if err != nil {
		entry.Error("history list", err)

		return
	}

	if len(h.Messages) == 0 {
		return
	}

	log.Infof("%d unread message(s) for '%s' label", len(h.Messages), v.labelID)

	for _, m := range h.Messages {
		var msg *message

		msg, err = p.loadMessage(m.Id)
		if err != nil {
			log.Errorf("load message [%s] %s", m.Id, err)

			continue
		}

		if msg.Body == "" {
			log.Errorf("empty message [%s] %s", m.Id, err)

			continue
		}

		fields := v.ParseSet.Parse(msg.Body)

		log.Println(msg.ID, "FROM", msg.From, "SUBJ", msg.Subject, "RES:", fields.Name, fields.Email, fields.Property)

		err = p.sendEventLink(ctx, entry, &fields, v.ListingPlatform)
		if err != nil {
			entry.Errorf("process [%s] %s", msg.ID, err)

			continue
		}

		_, err = p.GmailService.Users.Messages.Modify("me", msg.ID, &gmail.ModifyMessageRequest{
			RemoveLabelIds: []string{"UNREAD"},
		}).Do()
		if err != nil {
			entry.Errorf("modify [%s] %s", msg.ID, err)

			continue
		}
	}
}

func (p *Parser) sendEventLink(
	ctx context.Context,
	entry *log.Entry,
	fields *EmailFields,
	source mixpanel.ListingPlatform,
) error {
	prop, err := p.Connections.Listing.VersionGet(grpcctx.WithServiceUser(ctx), &listing.VersionGetRequest{
		Spec: listing.VersionSpecPropertyID("", fields.Property),
	})
	if err != nil {
		if errors.Is(err, grpcerr.NotFound) {
			return nil
		}

		entry.WithError(err).Error("couldn't fetch property")

		return err
	}

	err = p.Repository.EnquirySave(ctx, &model.Enquiry{
		PropertyID: prop.GetVersion().GetPropertyId(),
		ListingID:  prop.GetVersion().GetListingId(),
	})
	if err != nil {
		entry.WithError(err).Error("couldn't record enquiry")

		return err
	}

	if prop.GetVersion().GetFlags().GetTest().GetValue() {
		ctx = context.WithValue(ctx, grpcctx.CtxIsTestUser, true)
	}

	user, _ := p.getUserByEmail(ctx, prop.GetVersion(), fields.Email)

	_, err = p.Connections.Mixpanel.TrackEvent(ctx, &mixpanel.TrackRequest{
		UserId: user,
		Email:  fields.Email,
		Event: &mixpanel.TrackRequest_ViewingInvitation{
			ViewingInvitation: &mixpanel.Event_ViewingInvitation{
				Email:            fields.Email,
				PropertyId:       prop.GetVersion().GetPropertyId(),
				InvitationOrigin: source,
			},
		},
	})
	if err != nil {
		entry.WithError(err).Error("Tracking ViewingInvitation event")
	}

	authOrigin := "zoopla"

	if source == mixpanel.ListingPlatform_LISTING_PLATFORM_RIGHTMOVE {
		authOrigin = "rightmove"
	}

	_, err = p.Connections.Notification.SendNotification(ctx, &notification.SendNotificationRequest{
		Entity: notification.Entity_ENTITY_PROPERTY_QUESTION,
		Action: notification.Action_ACTION_CREATE,
		Targets: []*notification.Target{
			{
				User: user,
				Role: notification.Role_ROLE_TENANT,
				Contacts: &notification.Contacts{
					Email:    fields.Email,
					Phone:    fields.Phone,
					Fullname: fields.Name,
				},
			},
		},
		Embeds: []*notification.EventEmbed{
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_PROPERTY,
				EmbedValue: prop.GetVersion().GetPropertyId(),
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_USER,
				EmbedValue: user,
			},
			{
				EmbedType:  messaging.EmbedType_EMBED_TYPE_AUTH_ORIGIN,
				EmbedValue: authOrigin,
			},
		},
	})
	if err != nil {
		entry.WithError(err).Error("sending sms property link")
	}

	return nil
}

func (p *Parser) getUserByEmail(ctx context.Context, vers *listing.Version, email string) (string, error) {
	if vers == nil {
		return "", errNoProperty
	}

	if vers.GetPeople() == nil {
		return "", errNoUser
	}

	people := vers.GetPeople()

	if people.GetLandlord() != nil {
		acc, err := p.Connections.Account.Get(ctx, &account.GetRequest{
			Id:    people.GetLandlord().GetValue(),
			Admin: false,
		})
		if err == nil {
			if acc.Account.GetContact() != nil {
				if acc.Account.GetContactValue().GetEmailPrimaryValue() == email {
					return people.GetLandlord().GetValue(), nil
				}
			}
		}
	}

	if people.GetTenants() == nil {
		return "", errNoUser
	}

	for _, t := range people.GetTenants().GetValue() {
		acc, err := p.Connections.Account.Get(ctx, &account.GetRequest{
			Id:    t,
			Admin: false,
		})
		if err == nil {
			if acc.Account.GetContact() != nil {
				if acc.Account.GetContactValue().GetEmailPrimaryValue() == email {
					return t, nil
				}
			}
		}
	}

	return "", errNoUser
}

func getOneOfRegexps(content string, rs []*regexp.Regexp) string {
	for _, r := range rs {
		if v := r.FindStringSubmatch(content); len(v) > 1 {
			res := strings.TrimSpace(v[1])
			if res != "" {
				return res
			}
		}
	}

	return ""
}

func formatPhoneNumber(phoneNumber string) string {
	num, err := phonenumbers.Parse(phoneNumber, "GB")
	if err != nil {
		return ""
	}

	return phonenumbers.Format(num, phonenumbers.E164)
}

func (p *Parser) handleSubscription(ctx context.Context, v *Subscription) error {
	entry := log.WithField("source", v.labelID).WithField("subscription", v.Name)

	prof, err := p.GmailService.Users.GetProfile("me").Do()
	if err != nil {
		return err
	}

	last := prof.HistoryId

	cond := sync.NewCond(&sync.Mutex{})

	var ready bool

	go func() {
		for {
			cond.L.Lock()

			for !ready {
				cond.Wait()
			}

			ready = false

			cond.L.Unlock()

			p.processUnread(ctx, entry, v)
		}
	}()

	go func() {
		for range time.Tick(p.ScheduleInterval) {
			cond.Broadcast()
		}
	}()

	go func() {
		err = p.PubsubClient.Subscription(v.subscription).Receive(ctx, func(ctx context.Context, m *pubsub.Message) {
			var payload notificationPayload

			err = json.Unmarshal(m.Data, &payload)
			if err != nil {
				entry.Error(err)

				return
			}

			cond.L.Lock()

			ready = true

			cond.L.Unlock()

			cond.Broadcast()

			if payload.HistoryID > last {
				last = payload.HistoryID
			}

			m.Ack()
		})
		if err != nil {
			entry.Error(err)

			return
		}
	}()

	return nil
}

func (p *Parser) subscribe() error {
	ls, err := p.GmailService.Users.Labels.List("me").Do()
	if err != nil {
		return err
	}

	for _, l := range ls.Labels {
		t, ok := p.labels[l.Name]
		if !ok {
			continue
		}

		_, err = p.GmailService.Users.Watch("me", &gmail.WatchRequest{
			LabelIds:  []string{l.Id},
			TopicName: t.topic,
		}).Do()
		if err != nil {
			return err
		}

		t.labelID = l.Id
	}

	return nil
}

// Setup the subscriptions
func (p *Parser) Setup(ctx context.Context) error {
	err := p.subscribe()
	if err != nil {
		return err
	}

	for _, v := range p.Subscriptions {
		err = p.handleSubscription(ctx, v)
		if err != nil {
			return err
		}
	}

	return nil
}
